#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of CASCADe package
#
# Developed within the ExoplANETS-A H2020 program.
#
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018, 2020  Jeroen Bouwman
"""
This module defines the data models for the CASCADe transit spectroscopy code
"""

import numpy as np
import astropy.units as u


__all__ = ['SpectralData', 'SpectralDataTimeSeries',
           'MeasurementDesc', 'UnitDesc', 'FlagDesc',
           'AuxilaryInfoDesc']


class InstanceDescriptorMixin:
    """
    The InstanceDescriptorMixin Class.

    Mixin to be able to add descriptor to the instance of the class
    and not the class itself
    """

    def __getattribute__(self, name):
        value = object.__getattribute__(self, name)
        if hasattr(value, '__get__'):
            value = value.__get__(self, self.__class__)
        return value

    def __setattr__(self, name, value):
        try:
            obj = object.__getattribute__(self, name)
        except AttributeError:
            pass
        else:
            if hasattr(obj, '__set__'):
                return obj.__set__(self, value)
        return object.__setattr__(self, name, value)


class UnitDesc:
    """
    The UnitDesc Class.

    A descriptor for adding auxilary measurements,
    setting the property for the unit atribute
    """

    def __init__(self, keyname):
        self.default = None
        self.values = {}
        self.keyname = keyname

    def __get__(self, instance, owner):
        """
        Get.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        owner : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return getattr(instance, "_" + self.keyname, self.default)

    def __set__(self, instance, value):
        """
        Set.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        if hasattr(instance, "_" + self.keyname[:-5]):
            unit = getattr(instance, "_" + self.keyname, self.default)
            data = getattr(instance, "_" + self.keyname[:-5], np.array(0.0))
            if (unit is not None) and (value is not None):
                if data.shape == ():
                    setattr(instance, "_" + self.keyname[:-5],
                            ((np.array([data]) * unit).to(value)).value)
                else:
                    setattr(instance, "_" + self.keyname[:-5],
                            ((data * unit).to(value)).value)
        setattr(instance, "_" + self.keyname, value)

    def __delete__(self, instance):
        """
        Delete.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        delattr(instance, "_" + self.keyname)
        del self.values[instance]


class FlagDesc:
    """
    The FlagDesc Class.

    A descriptor for adding logical flags
    """

    def __init__(self, keyname):
        self.default = None
        self.values = {}
        self.keyname = keyname

    def __get__(self, instance, owner):
        """
        Get.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        owner : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return getattr(instance, "_" + self.keyname, None)

    def __set__(self, instance, value):
        """
        Set.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        setattr(instance, "_" + self.keyname, value)

    def __delete__(self, instance):
        """
        Delete.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        delattr(instance, "_" + self.keyname)
        del self.values[instance]


class AuxilaryInfoDesc:
    """
    The AuxilaryInfoDesc Class.

    A descriptor for adding Auxilary information to the dataset
    """

    def __init__(self, keyname):
        self.default = None
        self.values = {}
        self.keyname = keyname

    def __get__(self, instance, owner):
        """
        Get.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        owner : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return getattr(instance, "_" + self.keyname, None)

    def __set__(self, instance, value):
        """
        Set.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        setattr(instance, "_" + self.keyname, value)

    def __delete__(self, instance):
        """
        Delete.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        delattr(instance, "_" + self.keyname)
        del self.values[instance]


class MeasurementDesc:
    """
    The MeasurementDesc Class.

    A descriptor for adding auxilary measurements,
    setting the properties for the the measurement and unit
    """

    def __init__(self, keyname):
        self.default = float("NaN")
        self.values = {}
        self.keyname = keyname

    def __get__(self, instance, owner):
        """
        Get.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        owner : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        unit = getattr(instance, "_" + self.keyname + "_unit", None)
        if unit is not None:
            unit_out = unit
        else:
            unit_out = u.dimensionless_unscaled
        mask = getattr(instance, "_mask")
        value = getattr(instance, "_" + self.keyname, self.default)
        if (mask.shape == ()) or (mask.shape == value.shape):
            value_out = value
        else:
            ntile = len(value.shape)
            tiling = getattr(instance, "_data").shape[:-ntile] + \
                tuple(np.ones(ntile).astype(int))
            value_out = np.tile(value, tiling)
        return np.ma.array(value_out * unit_out, mask=mask)

    def __set__(self, instance, value):
        """
        Set.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        if isinstance(value, u.Quantity):
            unit = getattr(instance, "_" + self.keyname + "_unit", None)
            if unit is not None:
                setattr(instance, "_" + self.keyname,
                        np.array(((value).to(unit)).value))
            else:
                setattr(instance, "_" + self.keyname, np.array(value.value))
                setattr(instance, "_" + self.keyname + "_unit", value.unit)
        else:
            if np.array(value).shape != ():
                setattr(instance, "_" + self.keyname, np.array(value))
            else:
                setattr(instance, "_" + self.keyname, np.array([value]))

    def __delete__(self, instance):
        """
        Delete.

        Parameters
        ----------
        instance : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        delattr(instance, "_" + self.keyname)
        del self.values[instance]


class SpectralData(InstanceDescriptorMixin):
    """
    The SpectralData Class.

    Class defining basic properties of spectral data
    In the instance if the SpectralData class
    all data are stored internally as numppy arrays. Outputted data are
    astropy Quantities unless no units (=None) are specified.

    Parameters
    ----------
    wavelength
        wavelength of data (can be frequencies)
    wavelenth_unit
        The physical unit of the wavelength (uses astropy.units)
    data
        spectral data
    data_unit
        the physical unit of the data (uses astropy.units)
    uncertainty
        uncertainty on spectral data
    mask
        mask defining masked data
    **kwargs
        any auxilary data relevant to the spectral data
        (like position, detector temperature etc.)
        If unit is not explicitly given a unit atribute is added.
        Input argument can be instance of astropy quantity.
        Auxilary atributes are added to instance of the SpectralData class
        and not to the class itself. Only the required input stated above
        is always defined for all instances.

    Examples
    --------
    To create an instance of a SpectralData object with an
    initialization with data using units, run the following code:

    >>> import numpy as np
    >>> import astropu.units as u
    >>> from cascade.data_model import SpectralData

    >>> wave = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])*u.micron
    >>> flux = np.array([8.0, 8.0, 8.0, 8.0, 8.0, 8.0])*u.Jy
    >>> sd = SpectralData(wavelength=wave, data=flux)

    >>> print(sd.data, sd.wavelength)

    To change to convert the units to a different but equivalent unit:

    >>> sd.data_unit = u.erg/u.s/u.cm**2/u.Hz
    >>> sd.wavelength_unit = u.cm
    >>> print(sd.data, sd.wavelength)

    """

    def __init__(self, wavelength=float("NaN"), wavelength_unit=None,
                 data=float("NaN"), data_unit=None, uncertainty=float("NaN"),
                 mask=False, **kwargs):
        self._wavelength_unit = wavelength_unit  # Do not change init. order
        self.wavelength = wavelength
        self._data_unit = data_unit  # Do not change init. order
        self.data = data
        self.uncertainty = uncertainty
        if (not isinstance(data, np.ma.masked_array)) | \
                (np.array(mask).shape == np.array(data).shape):
            self.mask = mask
        # setting optional keyword parameters
        for key, value in kwargs.items():
            # check for unit argument
            if "_unit" in key:
                setattr(self, key, UnitDesc(key))
                setattr(self, "_" + key, value)
        for key, value in kwargs.items():
            # set auxilary data
            if "_unit" in key:
                continue
            # check if not boolean
            if isinstance(value, type(True)):
                setattr(self, key, FlagDesc(key))
                setattr(self, key, value)
                continue
            if not isinstance(value, (np.ndarray)):
                setattr(self, key, AuxilaryInfoDesc(key))
                setattr(self, key, value)
                continue
            if any(isinstance(t, str) for t in value):
                setattr(self, key, AuxilaryInfoDesc(key))
                setattr(self, key, value)
                continue
            if not hasattr(self, "_" + key + "_unit"):
                # if unit is not explicitly given, set property
                setattr(self, key + "_unit",
                        UnitDesc(key + "_unit"))
                setattr(self, "_" + key + "_unit", None)
            setattr(self, key, MeasurementDesc(key))
            setattr(self, key, value)

    @property
    def wavelength(self):
        """
        Wavelength.

        The wavelength atttribute of the SpectralData is defined
        through a getter and setter method. This ensures that the
        returned wavelength has always a unit associated with it (if the
        wavelength_unit is set) and that the returned  wavelength has the same
        dimension and mask as the data attribute.
        """
        if self._wavelength_unit is not None:
            unit_out = self._wavelength_unit
        else:
            unit_out = u.dimensionless_unscaled
        if (self.mask.shape == ()) or \
           (self.mask.shape == self._wavelength.shape):
            wavelength_out = self._wavelength
        else:
            ntile = len(self._wavelength.shape)
            tiling = ((self._data.shape)[::-1])[:-ntile] + \
                tuple(np.ones(ntile).astype(int))
            wavelength_out = np.tile(self._wavelength.T, tiling).T
        if wavelength_out.shape == ():
            return np.ma.array(np.array([wavelength_out]) * unit_out,
                               mask=self.mask)
        return np.ma.array(wavelength_out * unit_out, mask=self.mask)

    @wavelength.setter
    def wavelength(self, value):
        if isinstance(value, np.ma.masked_array):
            wave_in = value.data
            mask_in = np.ma.getmaskarray(value)
            self._mask = mask_in
        else:
            wave_in = value
        if isinstance(wave_in, u.Quantity):
            if self._wavelength_unit is not None:
                self._wavelength = \
                    np.array(((wave_in).to(self._wavelength_unit,
                                           equivalencies=u.spectral())).value)
            else:
                self._wavelength = np.array(wave_in.value)
                self._wavelength_unit = wave_in.unit
        else:
            if np.array(wave_in).shape != ():
                self._wavelength = np.array(wave_in)
            else:
                self._wavelength = np.array([wave_in])

    @property
    def wavelength_unit(self):
        """
        Wavelength Unit.

        The wavelength_unit attribute of the SpectralData is defined
        through a getter and setter method. This ensures that units can be
        updated and the wavelength value will be adjusted accordingly.
        """
        return self._wavelength_unit

    @wavelength_unit.setter
    def wavelength_unit(self, value):
        if hasattr(self, '_wavelength'):
            if (self._wavelength_unit is not None) and (value is not None):
                if self._wavelength.shape == ():
                    self._wavelength = \
                        ((np.array([self._wavelength]) *
                          self._wavelength_unit).
                         to(value, equivalencies=u.spectral())).value
                else:
                    self._wavelength = \
                        ((self._wavelength * self._wavelength_unit).
                         to(value, equivalencies=u.spectral())).value
        self._wavelength_unit = value

    @property
    def data(self):
        """
        Definition of the data atttribute of the SpectralData.

        The data atttribute of the SpectralData is defined through a
        getter and setter method. In case data is initialized with a
        masked quantity, the data_unit and mask attributes will be set
        automatically.
        """
        if self._data_unit is not None:
            unit_out = self._data_unit
        else:
            unit_out = u.dimensionless_unscaled
        if self._data.shape == ():
            return np.ma.array(np.array([self._data]) * unit_out,
                               mask=self.mask)
        return np.ma.array(self._data * unit_out, mask=self.mask)

    @data.setter
    def data(self, value):
        if isinstance(value, np.ma.masked_array):
            data_in = value.data
            mask_in = np.ma.getmaskarray(value)
            self._mask = mask_in
        else:
            data_in = value
        if isinstance(data_in, u.Quantity):
            if self._data_unit is not None:
                self._data = np.array(((data_in).to(self._data_unit)).value)
            else:
                self._data = np.array(data_in.value)
                self._data_unit = data_in.unit
        else:
            if np.array(data_in).shape != ():
                self._data = np.array(data_in)
            else:
                self._data = np.array([data_in])

    @property
    def uncertainty(self):
        """
        Uncertainty.

        The uncertainty atttribute of the SpectralData is defined
        through a getter and setter method. This ensures that the
        returned uncertainty has the same unit associated with it
        (if the data_unit is set) and the same mask as the data attribute.
        """
        if self._data_unit is not None:
            unit_out = self._data_unit
        else:
            unit_out = u.dimensionless_unscaled
        if self._uncertainty.shape == ():
            return np.ma.array(np.array([self._uncertainty]) * unit_out,
                               mask=self.mask)
        if np.all(np.isnan(self._uncertainty)):
            return np.ma.array(self._uncertainty * unit_out)
        return np.ma.array(self._uncertainty * unit_out, mask=self.mask)

    @uncertainty.setter
    def uncertainty(self, value):
        if isinstance(value, np.ma.masked_array):
            data_in = value.data
            mask_in = np.ma.getmaskarray(value)
            self._mask = mask_in
        else:
            data_in = value
        if isinstance(data_in, u.Quantity):
            if self._data_unit is not None:
                self._uncertainty = \
                    np.array(((data_in).to(self._data_unit)).value)
            else:
                self._uncertainty = np.array(data_in.value)
                self._data_unit = data_in.unit
        else:
            if np.array(data_in).shape != ():
                self._uncertainty = np.array(data_in)
            else:
                self._uncertainty = np.array([data_in])

    @property
    def data_unit(self):
        """
        Definition of the data unit attribute of the SpectralData.

        The data_unit attribute of the SpectralData is defined
        through a getter and setter method. This ensures that units can be
        updated and the data value will be adjusted accordingly.
        """
        return self._data_unit

    @data_unit.setter
    def data_unit(self, value):
        if hasattr(self, '_data'):
            if (value is not None) and (self._data_unit is not None):
                if self._data.shape == ():
                    self._data = \
                        ((np.array([self._data]) * self._data_unit).
                         to(value)).value
                else:
                    self._data = \
                        ((self._data * self._data_unit).to(value)).value
                if self._uncertainty.shape == ():
                    self._uncertainty = \
                        ((np.array([self._uncertainty]) * self._data_unit).
                         to(value)).value
                else:
                    self._uncertainty = \
                        ((self._uncertainty * self._data_unit).to(value)).value
        self._data_unit = value

    @property
    def mask(self):
        """
        Mask.

        The mask atttribute of the SpectralData is defined
        through a getter and setter method. This ensures that the
        returned mask has the same dimension as the data attribute and
        will be set automatically if the input data is a masked array.
        """
        if self._mask.shape == ():
            return np.array([self._mask])
        return self._mask

    @mask.setter
    def mask(self, value):
        if np.array(value).shape == ():
            self._mask = np.array([value]).astype(bool)
        else:
            self._mask = np.array(value).astype(bool)

    def return_masked_array(self, attr):
        """
        Return the data, wavelength or uncertainty as masked array.

        Parameters
        ----------
        attr : 'str'
            String value to indicate which measurement should be returned as
            masked array. Can be 'data', 'wavelength' or 'uncertainty'.

        Returns
        -------
        'np.ma.maskedArray'
            The measurement corresponding to the atts value as masked array.

        """
        if attr == 'data':
            if self._data.shape == ():
                return np.ma.array(np.array([self._data]), mask=self.mask)
            return np.ma.array(self._data, mask=self.mask)
        elif attr == 'wavelength':
            if (self.mask.shape == ()) or \
               (self.mask.shape == self._wavelength.shape):
                wavelength_out = self._wavelength
            else:
                ntile = len(self._wavelength.shape)
                tiling = ((self._data.shape)[::-1])[:-ntile] + \
                    tuple(np.ones(ntile).astype(int))
                wavelength_out = np.tile(self._wavelength.T, tiling).T
            if wavelength_out.shape == ():
                return np.ma.array(np.array([wavelength_out]), mask=self.mask)
            return np.ma.array(wavelength_out, mask=self.mask)
        elif attr == 'uncertainty':
            if self._uncertainty.shape == ():
                return np.ma.array(np.array([self._uncertainty]),
                                   mask=self.mask)
            if np.all(np.isnan(self._uncertainty)):
                return np.ma.array(self._uncertainty)
            return np.ma.array(self._uncertainty, mask=self.mask)

    def add_measurement(self, **kwargs):
        """
        Add measuement to SpectralData object.

        Parameters
        ----------
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        # setting optional keyword parameters
        for key, value in kwargs.items():
            # check for unit argument
            if "_unit" in key:
                setattr(self, key, UnitDesc(key))
                setattr(self, "_" + key, value)
        for key, value in kwargs.items():
            # set auxilary data
            if "_unit" not in key:
                if not hasattr(self, "_" + key + "_unit"):
                    # if unit is not explicitly given, set property
                    setattr(self, key + "_unit",
                            UnitDesc(key + "_unit"))
                    setattr(self, "_" + key + "_unit", None)
                setattr(self, key, MeasurementDesc(key))
                setattr(self, key, value)

    def add_flag(self, **kwargs):
        """
        Add flag to the SpectralData object.

        Parameters
        ----------
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        for key, value in kwargs.items():
            if isinstance(value, type(True)):
                setattr(self, key, FlagDesc(key))
                setattr(self, key, value)

    def add_auxilary(self, **kwargs):
        """
        Add auxilary data to the SpectralData object.

        Parameters
        ----------
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        for key, value in kwargs.items():
            setattr(self, key, AuxilaryInfoDesc(key))
            setattr(self, key, value)


class SpectralDataTimeSeries(SpectralData):
    """
    The Spectral Data TimeSeries Class.

    Class defining timeseries of spectral data. This class inherits from
    SpectralData. The data stored within this class has one additional
    time dimension

    Parameters
    ----------
    wavelength : 'array_like'
        wavelength assigned to each data point (can be also be frequencies)
    wavelenth_unit : 'astropy.units.core.Unit'
        The physical unit of the wavelength .
    data : 'array_like'
        The spectral data to be analysed. This can be either spectra (1D),
        spectral images (2D) or spectral data cubes (3D).
    data_unit : astropy.units.core.Unit
        The physical unit of the data.
    uncertainty
        The uncertainty associated with the spectral data.
    mask : 'array_like'
        The bad pixel mask flagging all data not to be used.
    **kwargs
        any auxilary data relevant to the spectral data
        (like position, detector temperature etc.)
        If unit is not explicitly given a unit atribute is added.
        Input argument can be instance of astropy quantity.
        Auxilary atributes are added to instance of the SpectralData class
        and not to the class itself. Only the required input stated above
        is always defined for all instances.
    time : 'array_like'
        The time of observation assiciated with each data point.
    time_unit : 'astropy.units.core.Unit'
        physical unit of time data

    Examples
    --------
    To create an instance of a SpectralDataaTimeSeries object with an
    initialization with data using units, run the following code:

    >>> import numpy as np
    >>> from cascade.data_model import SpectralDataTimeSeries

    >>> wave = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])*u.micron
    >>> flux = np.array([8.0, 8.0, 8.0, 8.0, 8.0, 8.0])*u.Jy
    >>> time = np.array([240000.0, 2400001.0, 2400002.0])*u.day
    >>> flux_time_series = np.repeat(flux[:, np.newaxis], time.shape[0], 1)
    >>> sdt = SpectralDataTimeSeries(wavelength=wave, data=flux_time_series,
                                     time=time)

    """

    def __init__(self, wavelength=float("NaN"), wavelength_unit=None,
                 data=np.array([[float("NaN")]]), data_unit=None,
                 uncertainty=np.array([[float("NaN")]]), mask=False,
                 time=float("NaN"), time_unit=None, **kwargs):
        self._time_unit = time_unit  # Do not change order
        self.time = time
        super(SpectralDataTimeSeries,
              self).__init__(wavelength=wavelength,
                             wavelength_unit=wavelength_unit,
                             data=data, data_unit=data_unit,
                             uncertainty=uncertainty, mask=mask, **kwargs)

    @property
    def time(self):
        """
        Time.

        The time atttribute of the SpectralDataTimeSeries is defined
        through a getter and setter method. This ensures that the
        returned time has always a unit associated with it if the time_unit is
        set and that the returned time has the same dimension and mask as
        the data attribute.
        """
        if self._time_unit is not None:
            unit_out = self._time_unit
        else:
            unit_out = u.dimensionless_unscaled
        if (self.mask.shape == ()) or (self.mask.shape == self._time.shape):
            time_out = self._time
        else:
            ntile = len(self._time.shape)
            tiling = (self._data.shape)[:-ntile] + \
                tuple(np.ones(ntile).astype(int))
            time_out = np.tile(self._time, tiling)
        if time_out.shape == ():
            return np.ma.array(np.array([time_out]) * unit_out,
                               mask=self.mask)
        else:
            return np.ma.array(time_out * unit_out, mask=self.mask)

    @time.setter
    def time(self, value):
        if isinstance(value, u.Quantity):
            if self._time_unit is not None:
                self._time = np.array(((value).to(self._time_unit)).value)
            else:
                self._time = np.array(value.value)
                self._time_unit = value.unit
        else:
            if np.array(value).shape != ():
                self._time = np.array(value)
            else:
                self._time = np.array([value])

    @property
    def time_unit(self):
        """
        Time Unit.

        The time_unit attribute of the SpectralDataTimeSeries is defined
        through a getter and setter method. This ensures that units can be
        updated and the time value will be adjusted accordingly.
        """
        return self._time_unit

    @time_unit.setter
    def time_unit(self, value):
        if hasattr(self, '_time'):
            if (value is not None) and (self._time_unit is not None):
                if self._time.shape == ():
                    self._time = \
                        ((np.array([self._time]) * self._time_unit).
                         to(value)).value
                else:
                    self._time = \
                        ((self._time * self._time_unit).to(value)).value
        self._time_unit = value

    def return_masked_array(self, attr):
        """
        Return a maasked array.

        Parameters
        ----------
        attr : TYPE
            DESCRIPTION.

        Raises
        ------
        ValueError
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if attr == 'time':
            if (self.mask.shape == ()) or \
               (self.mask.shape == self._time.shape):
                time_out = self._time
            else:
                ntile = len(self._time.shape)
                if ntile != 1:
                    raise ValueError("Unknown time structure")
                tiling = (self._data.shape[:-1]) + \
                    tuple(np.ones(ntile).astype(int))
                time_out = np.tile(self._time, tiling)
            if time_out.shape == ():
                return np.ma.array(np.array([time_out]), mask=self.mask)
            return np.ma.array(time_out, mask=self.mask)
        if attr == 'time_bjd':
            if (self.mask.shape == ()) or \
               (self.mask.shape == self._time_bjd.shape):
                time_out = self._time_bjd
            else:
                ntile = len(self._time_bjd.shape)
                if ntile != 1:
                    raise ValueError("Unknown time structure")
                tiling = (self._data.shape[:-1]) + \
                    tuple(np.ones(ntile).astype(int))
                time_out = np.tile(self._time_bjd, tiling)
            if time_out.shape == ():
                return np.ma.array(np.array([time_out]), mask=self.mask)
            return np.ma.array(time_out, mask=self.mask)
        elif attr == 'position':
            if hasattr(self, 'position'):
                if (self.mask.shape == ()) or \
                   (self.mask.shape == self._position.shape):
                    position_out = self._position
                else:
                    ntile = len(self._position.shape)
                    if ntile != 1:
                        raise ValueError("Unknown time structure")
                    tiling = (self._data.shape[:-1]) + \
                        tuple(np.ones(ntile).astype(int))
                    position_out = np.tile(self._position, tiling)
                if position_out.shape == ():
                    return np.ma.array(np.array([position_out]),
                                       mask=self.mask)
                return np.ma.array(position_out, mask=self.mask)
        elif attr == 'fwhm':
            if hasattr(self, 'fwhm'):
                if (self.mask.shape == ()) or \
                   (self.mask.shape == self._fwhm.shape):
                    fwhm_out = self._fwhm
                else:
                    ntile = len(self._fwhm.shape)
                    if ntile != 1:
                        raise ValueError("Unknown time structure")
                    tiling = (self._data.shape[:-1]) + \
                        tuple(np.ones(ntile).astype(int))
                    fwhm_out = np.tile(self._fwhm, tiling)
                if fwhm_out.shape == ():
                    return np.ma.array(np.array([fwhm_out]),
                                       mask=self.mask)
                return np.ma.array(fwhm_out, mask=self.mask)
        elif attr == 'scaling':
            if hasattr(self, 'scaling'):
                if (self.mask.shape == ()) or \
                   (self.mask.shape == self._scaling.shape):
                    scaling_out = self._scaling
                else:
                    ntile = len(self._scaling.shape)
                    if ntile != 1:
                        raise ValueError("Unknown time structure")
                    tiling = (self._data.shape[:-1]) + \
                        tuple(np.ones(ntile).astype(int))
                    scaling_out = np.tile(self._scaling, tiling)
                if scaling_out.shape == ():
                    return np.ma.array(np.array([scaling_out]),
                                       mask=self.mask)
                return np.ma.array(scaling_out, mask=self.mask)
        return super().return_masked_array(attr)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of the CASCADe package which has been
# developed within the ExoplANETS-A H2020 program.
#
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2020  Jeroen Bouwman
"""
Created on April 24 2020

@author:Jeroen Bouwman, Rene Gastaud, Raphael Peralta, Fred Lahuis
"""
import os
import pickle
import click
import sys
import six
from pyfiglet import figlet_format
from click import Option, UsageError
import time


try:
    import colorama
    colorama.init()
except ImportError:
    colorama = None

try:
    from termcolor import colored
except ImportError:
    colored = None

__all__ = ['built_local_hst_archive']


def log(string, color, font="slant", figlet=False):
    if colored:
        if not figlet:
            six.print_(colored(string, color))
        else:
            six.print_(colored(figlet_format(
                string, font=font), color))
    else:
        six.print_(string)


class MutuallyExclusiveOption(Option):
    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop('mutually_exclusive', []))
        help = kwargs.get('help', '')
        if self.mutually_exclusive:
            ex_str = ', '.join(self.mutually_exclusive)
            kwargs['help'] = help + (
                ' NOTE: This argument is mutually exclusive with '
                ' arguments: [' + ex_str + '].'
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(
                    self.name,
                    ', '.join(self.mutually_exclusive)
                )
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(
            ctx,
            opts,
            args
        )


def check_path_option(new_path, environent_variable, message):
    if new_path is not None:
        if os.path.isabs(new_path):
            os.environ[environent_variable] = new_path
        else:
            try:
                base_new_path = os.environ[environent_variable]
                os.environ[environent_variable] = \
                    os.path.join(base_new_path, new_path)
            except KeyError:
                log(message, "red")
                sys.exit()


@click.command()
@click.option('--init_path',
              '-ip',
              nargs=1,
              type=click.STRING,
              help='Path to the directory containing the configuration files.'
                   'If not specified, the value set by the environment '
                   'variable CASCADE_INITIALIZATION_FILE_PATH is used or if '
                   'neither is set it defaults to the CASCADe default value '
                   'of the CASCADe distribution. This path setting can be '
                   'relative to the absolute path given by the '
                   'CASCADE_INITIALIZATION_FILE_PATH environment variable.',
              )
@click.option('--data_path',
              '-dp',
              nargs=1,
              type=click.STRING,
              help='Path to the data (observations, calibration files, '
                   'exoplanet catalogs, archive databases) needed by CASCADe '
                   'to function. If not set, the value set by the environment '
                   'variable CASCADE_DATA_PATH is used or if neither is set '
                   'it defaults to the CASCADe default value of the CASCADe '
                   'distribution.',
              )
@click.option('--save_path',
              '-sp',
              nargs=1,
              type=click.STRING,
              help='Path to the directory where CASCADe saves results '
                   '(plots, extracted planetary spectra). If not specified, '
                   'the value set by the environment variable '
                   'CASCADE_SAVE_PATH is used, or if neither is set it '
                   'defaults to the CASCADe default value of the CASCADe '
                   'distribution.',
              )
@click.option('--no_warnings',
              '-nw',
              is_flag=True,
              default=False,
              help='If set no warning messages are printed to stdev. '
                   'Default is False.',
              )
@click.option('--primary_exoplanet_catalog',
              '-pc',
              nargs=1,
              type=click.STRING,
              default='NASAEXOPLANETARCHIVE',
              help='The name of the primary exoplanet catalog used to create '
                   'the object.ini file.'
              )
@click.option('--list_all_planets',
              '-lap', cls=MutuallyExclusiveOption,
              is_flag=True,
              default=False,
              help='Return all exoplanet names in Archive file.',
              mutually_exclusive=['list_catalog_id', 'visits',
                                  'all_visits_planet', 'download_all_data'],
              )
@click.option('--list_catalog_id',
              '-lci', cls=MutuallyExclusiveOption,
              nargs=1,
              type=click.STRING,
              help='Returns all observation calatog id of the planet which '
                   'name is given as an argument.',
              mutually_exclusive=['list_all_planets', 'visits',
                                  'all_visits_planet', 'download_all_data']
              )
@click.option('--visits',
              '-v', cls=MutuallyExclusiveOption,
              multiple=True,
              help="All initialization files will be created and the archive "
                   "data will be downloaded for the system of which the "
                   "calatog id is given as an argument.",
              mutually_exclusive=['list_all_planets', 'list_catalog_id',
                                  'all_visits_planet', 'download_all_data']
              )
@click.option('--all_visits_planet',
              '-avp', cls=MutuallyExclusiveOption,
              nargs=1,
              type=click.STRING,
              help="All initialization files will be created and the "
                   "archive data will be downloaded for all visits "
                   "of the planet given as an argument.",
              mutually_exclusive=['list_all_planets', 'list_catalog_id',
                                  'visits', 'download_all_data']
              )
@click.option('--download_all_data',
              '-dad', cls=MutuallyExclusiveOption,
              is_flag=True,
              default=False,
              help='Download all the archived observations to local disk.',
              mutually_exclusive=['list_all_planets', 'list_catalog_id',
                                  'visits', 'all_visits_planet'],
              )
@click.option('--create_ini_files_only',
              '-cio',
              is_flag=True,
              help='If set only the ini files are created. '
                   'Default is False.'
              )
@click.option('--skip_existing_data',
              '-sed',
              is_flag=True,
              default=False,
              help='If set already downloaded data is skipped. '
                   'Default is False.'
              )
@click.option('--skip_existing_initalization_files',
              '-sei',
              is_flag=True,
              default=False,
              help='If set already created initialization files are skipped. '
                   'Default is False.'
              )
def built_local_hst_archive(init_path, data_path, save_path, no_warnings,
                            primary_exoplanet_catalog, list_all_planets,
                            list_catalog_id, visits, all_visits_planet,
                            download_all_data, create_ini_files_only,
                            skip_existing_data,
                            skip_existing_initalization_files):
    """
    Build local HST archive.

    This function creates a local copy of exoplanet observations in the
    MAST archive.
    """
    if no_warnings:
        os.environ["CASCADE_WARNINGS"] = 'off'
    else:
        os.environ["CASCADE_WARNINGS"] = 'on'

    if init_path is not None:
        check_path_option(init_path, "CASCADE_INITIALIZATION_FILE_PATH",
                          "Relative init_path given without setting the "
                          "CASCADE_INITIALIZATION_FILE_PATH environment "
                          "variable Stopping script")

    if data_path is not None:
        check_path_option(data_path, "CASCADE_DATA_PATH",
                          "Relative data_path given without setting the "
                          "CASCADE_DATA_PATH environment "
                          "variable Stopping script")

    if save_path is not None:
        check_path_option(save_path, "CASCADE_SAVE_PATH",
                          "Relative save_path given without setting the "
                          "CASCADE_SAVE_PATH environment "
                          "variable Stopping script")

    import cascade
    from cascade.initialize import cascade_default_data_path
    from cascade.initialize import cascade_default_path
    from cascade.initialize import cascade_default_initialization_path
    from cascade.initialize import cascade_default_save_path
    from cascade.initialize import cascade_default_scripts_path
    from cascade.build_archive import return_exoplanet_catalogs
    from cascade.build_archive import return_all_hst_planets
    from cascade.build_archive import return_hst_data_calalog_keys
    # from cascade.build_archive import long_substr
    from cascade.build_archive import return_header_info
    from cascade.build_archive import save_observations
    from cascade.build_archive import IniFileParser
    from cascade.build_archive import create_bash_script
    from cascade.build_archive import check_for_exceptions
    from cascade.build_archive import create_unique_ids

    log('CASCADe', color="blue", figlet=True)
    log("version {}, Copyright (C) 2020 "
        "EXOPANETS_A H2020 Program".format(cascade.__version__), "blue")
    log("The initialization file directory is set to: "
        "{}".format(cascade_default_initialization_path), "green")
    log("The data directory is set to: "
        "{}".format(cascade_default_data_path), "green")
    log("The save directory is set to: "
        "{}".format(cascade_default_save_path), "green")
    log("The scripts directory is set to: "
        "{}".format(cascade_default_scripts_path), "green")  

    # Location of the tamples
    TEMPLATES_DIR = os.path.join(cascade_default_path,
                                 'configuration_templates/')

    # All configuration files for the different sections in the
    # initialization files
    OBJECT_CONFIGURATION_FILE = 'object.conf'
    DILUTION_CONFIGURATION_FILE = 'dilution.conf'
    CATALOG_CONFIGURATION_FILE = 'catalog.conf'
    CASCADE_CONFIGURATION_FILE = 'cascade.conf'
    PROCESSING_CONFIGURATION_FILE = 'processing.conf'
    INSTRUMENT_CONFIGURATION_FILE = 'instrument.conf'
    OBSERVATIONS_CONFIGURATION_FILE = 'observations.conf'
    CPM_CONFIGURATION_FILE = 'cpm.conf'
    MODEL_CONFIGURATION_FILE = 'model.conf'

    # All initialization filetemplates
    OBJECT_CONFIGURATION_TEMPLATE = 'cascade_object_template.ini'
    EXTRACT_TIMESERIES_CONFIGURATION_TEMPLATE = \
        'cascade_extract_timeseries_template.ini'
    CALIBRATE_PLANET_SPECTRUM_CONFIGURATION_TEMPLATE = \
        'cascade_calibrate_planet_spectrum_template.ini'

    # In case of non standart issues with observations
    # list them in an exceptions.ini file
    INITIALIZATION_EXCEPTIONS = 'processing_exceptions.ini'
    INITIALIZATION_EXCEPTIONS_USER = 'user_processing_exceptions.ini'

    # Get the HST observations catalog file
    HST_CATALOG_FILE = os.path.join(cascade_default_path,
                                    "archive_databases/HST/WFC3/",
                                    "WFC3_files.pickle")
    with open(HST_CATALOG_FILE, 'rb') as f:
        hst_data_catalog = pickle.load(f)
    if list_all_planets:
        log("The observations catalog contains data for the following "
            "planets: \n {}".format(return_all_hst_planets(hst_data_catalog)),
            'green')
        sys.exit()
    if list_catalog_id is not None:
        log("For planet {}, the following observations catalog keys have been "
            "found: {}".format(list_catalog_id,
                               return_hst_data_calalog_keys(list_catalog_id,
                                                            hst_data_catalog)),
            'green')
        sys.exit()

    if all_visits_planet is not None:
        visits = return_hst_data_calalog_keys(all_visits_planet,
                                              hst_data_catalog)
        log("For planet {}, the following observations catalog keys have been "
            "found: {}".format(all_visits_planet, visits),
            'green')

    if (visits is None) & (all_visits_planet is None) & \
            (not download_all_data):
        log("Warning, neither visits and all_visits_planet are defined "
            "either speciefy on of the other of use the --all_data option "
            "to download the entire archive.", "red")
        sys.exit()

    if download_all_data:
        if click.confirm("Selected --download_all_data option. Do you want "
                         "to continue downloading the entier archive?"):
            visits = hst_data_catalog.keys()
        else:
            log("Aborting downloading the entier archive", "red")
            sys.exit()

    if len(visits) == 0:
        log("Warning, no visits found, check search or path settings", "red")
        sys.exit()

    # get all unique ID's and check for duplicates
    OBS_ID_DICT = create_unique_ids(hst_data_catalog)

    # explanet data catalogs
    catalogs_dictionary = return_exoplanet_catalogs()

    # ########### HERE LOOP STARTS #####################
    for visit in visits:
        PLANET_NAME = hst_data_catalog[visit]['planet']
        log("Target: {}, Visit: {}".format(PLANET_NAME, visit), 'green')

        # ############  some logic  ##############
        # ## Define the observation typpe and skips loop if problem ##
        if hst_data_catalog[visit]['observation'] == 'transit':
            OBS_TYPE = 'TRANSIT'
        elif hst_data_catalog[visit]['observation'] == 'eclipse':
            OBS_TYPE = 'ECLIPSE'
        else:
            OBS_TYPE = 'PROBLEM'
        if OBS_TYPE == 'PROBLEM':
            log('Skipping : {}'.format(hst_data_catalog[visit]['observation']),
                'red')
            continue
        # get the data files from hst databes
        data_files = hst_data_catalog[visit]['observations_id_ima'].split(',')
        cal_data_files = hst_data_catalog[visit]['calibrations_id'].split(',')
#        data_file_id = [file.split('_')[0] for file in data_files]
#        cal_data_file_id = [file.split('_')[0] for file in cal_data_files]

        # get or calculate all parameters needed for the observations and
        # instrument sections in the .ini file
#        OBS_ID = long_substr(data_file_id)

        OBS_ID = OBS_ID_DICT[PLANET_NAME][visit]['obs_id']
        OBS_ID_DIR = OBS_ID_DICT[PLANET_NAME][visit]['obs_id_dir']

        # get all needed parametesr from the data files and catalog files
        # to create the ini files for creating the spctral timeseries
        create_timeseries_namespace_dict = {}
        # fill dictionary with parameters to be filled into templates
        create_timeseries_namespace_dict['cascade_save_path'] = \
            PLANET_NAME+'_'+OBS_ID_DIR
        create_timeseries_namespace_dict['observations_type'] = OBS_TYPE
        create_timeseries_namespace_dict['observations_target_name'] = \
            PLANET_NAME+'_'+OBS_ID_DIR
        create_timeseries_namespace_dict['observations_id'] = OBS_ID
        create_timeseries_namespace_dict.update(
            **return_header_info(data_files[0], cal_data_files[0])
            )

        # update parameters for timeseries of 1D spectra to create the
        # ini files for the calibration of the planet specrum
        cal_planet_spec_namespace_dict = \
            create_timeseries_namespace_dict.copy()
        cal_planet_spec_namespace_dict['processing_nextraction'] = 1
        cal_planet_spec_namespace_dict['observations_data_product'] = 'COE'
        cal_planet_spec_namespace_dict['observations_has_background'] = False
        cal_planet_spec_namespace_dict['observations_mode'] = 'STARING'
        cal_planet_spec_namespace_dict['observations_data'] = 'SPECTRUM'

        # The name space for creating the ini files defining the object
        object_namespace_dict = {}

        # Check for exceptions on the standard values defined for a
        # particular system.
        exceptions_dict = \
            check_for_exceptions(INITIALIZATION_EXCEPTIONS,
                                 create_timeseries_namespace_dict)
        exceptions_dict_user = \
            check_for_exceptions(INITIALIZATION_EXCEPTIONS_USER,
                                 create_timeseries_namespace_dict)
        exceptions_dict.update(exceptions_dict_user)

        # ################ CREATE OBJECT.INI ##############################
        # update with exceptions
        object_namespace_dict.update(exceptions_dict)

        # create list with the configuration files to use
        configuration_file_list = [OBJECT_CONFIGURATION_FILE,
                                   CATALOG_CONFIGURATION_FILE,
                                   DILUTION_CONFIGURATION_FILE]

        # create initialization file parser object
        IFP = \
            IniFileParser(configuration_file_list,
                          OBJECT_CONFIGURATION_TEMPLATE,
                          object_namespace_dict,
                          TEMPLATES_DIR,
                          planet_name=PLANET_NAME,
                          exoplanet_catalogs_dictionary=catalogs_dictionary,
                          primary_exoplanet_catalog=primary_exoplanet_catalog)
        IFP.print_parser()
        object_parser = IFP.return_parser()

        # ################ CREATE EXTRACT_TIMESERIES.INI ####################

        # update with exceptions
        create_timeseries_namespace_dict.update(exceptions_dict)
        # create list with the configuration files to use
        configuration_file_list = [CASCADE_CONFIGURATION_FILE,
                                   PROCESSING_CONFIGURATION_FILE,
                                   MODEL_CONFIGURATION_FILE,
                                   INSTRUMENT_CONFIGURATION_FILE,
                                   OBSERVATIONS_CONFIGURATION_FILE]
        # create initialization file parser object
        IFP = \
            IniFileParser(configuration_file_list,
                          EXTRACT_TIMESERIES_CONFIGURATION_TEMPLATE,
                          create_timeseries_namespace_dict,
                          TEMPLATES_DIR)
        IFP.print_parser()
        extract_timeseries_parser = IFP.return_parser()

        # ############### CREATE CALIBRATE_PLANET_SPECTRUM.INI ################

        # update with exceptions
        cal_planet_spec_namespace_dict.update(exceptions_dict)

        # update which configureation files to use
        configuration_file_list = [CASCADE_CONFIGURATION_FILE,
                                   PROCESSING_CONFIGURATION_FILE,
                                   CPM_CONFIGURATION_FILE,
                                   MODEL_CONFIGURATION_FILE,
                                   INSTRUMENT_CONFIGURATION_FILE,
                                   OBSERVATIONS_CONFIGURATION_FILE]
        # create initialization file parser object
        IFP = \
            IniFileParser(configuration_file_list,
                          CALIBRATE_PLANET_SPECTRUM_CONFIGURATION_TEMPLATE,
                          cal_planet_spec_namespace_dict,
                          TEMPLATES_DIR)
        IFP.print_parser()
        calibrate_planet_spectrum_parser = IFP.return_parser()

        # ############## Saving ini files ##################
        initialization_save_path = os.path.join(
            cascade_default_initialization_path,
            extract_timeseries_parser['INSTRUMENT']["instrument_observatory"],
            extract_timeseries_parser['INSTRUMENT']["instrument"],
            extract_timeseries_parser['OBSERVATIONS']['observations_target_name']
            )
        os.makedirs(initialization_save_path, exist_ok=True)

        configuration_base_filename = \
            (
             "cascade_" +
             extract_timeseries_parser['OBSERVATIONS']
             ['observations_target_name']
             )
        configuration_object_filename = \
            configuration_base_filename+"_object.ini"
        if not (skip_existing_initalization_files & os.path.isfile(
            os.path.join(initialization_save_path,
                         configuration_object_filename))):
            with open(os.path.join(initialization_save_path,
                                   configuration_object_filename),
                      'w') as configfile:
                object_parser.write(configfile)

        configuration_extract_timeseries_filename = \
            configuration_base_filename+"_extract_timeseries.ini"

        if not (skip_existing_initalization_files & os.path.isfile(
            os.path.join(initialization_save_path,
                         configuration_extract_timeseries_filename))):
            with open(os.path.join(
                    initialization_save_path,
                    configuration_extract_timeseries_filename),
                    'w') as configfile:
                extract_timeseries_parser.write(configfile)

        configuration_calibrate_planet_spectrum_filename = \
            configuration_base_filename+"_calibrate_planet_spectrum.ini"
        if not (skip_existing_initalization_files & os.path.isfile(
            os.path.join(initialization_save_path,
                         configuration_calibrate_planet_spectrum_filename))):
            with open(os.path.join(
                    initialization_save_path,
                    configuration_calibrate_planet_spectrum_filename),
                    'w') as configfile:
                calibrate_planet_spectrum_parser.write(configfile)

        # # ################# GET ARCHIVE DATA ######################
        if not create_ini_files_only:
            save_observations(data_files, cal_data_files,
                              extract_timeseries_parser,
                              skip_existing=skip_existing_data)
        else:
            log("create_ini_files_only flag is True, not downloading data",
                "red")

        # ###################### crete bash script #######################
        create_bash_script(visit, extract_timeseries_parser)


if __name__ == '__main__':
    start_time = time.time()
    built_local_hst_archive()
    elapsed_time = time.time() - start_time
    log('elapsed time: {}'.format(elapsed_time), "green")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of CASCADe package
#
# Developed within the ExoplANETS-A H2020 program.
#
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018, 2021  Jeroen Bouwman
"""Module defiing utility functions used in cascade."""

import os
import glob
import fnmatch
import copy
import warnings
import numpy as np
from astropy.io import fits
from astropy.table import QTable
import astropy.units as u
from tqdm import tqdm
import numba as nb

from cascade.data_model import SpectralDataTimeSeries

__all__ = ['write_timeseries_to_fits', 'find', 'get_data_from_fits',
           'spectres', 'write_spectra_to_fits', '_define_band_limits',
           '_define_rebin_weights', '_rebin_spectra',
           'write_dataset_to_fits', 'read_dataset_from_fits']


def write_spectra_to_fits(spectral_dataset, path, filename, header_meta,
                          column_names=['Wavelength', 'Depth', 'Error Depth']):
    """
    Write spectra dataset object to fits files.

    Parameters
    ----------
    data : 'ndarray' or 'SpectralDataTimeSeries'
        The data cube which will be save to fits file. For each time step
        a fits file will be generated.
    path : 'str'
        Path to the directory where the fits files will be saved.
    filename: 'str' (optional)
        file name of save fits file.
    header_meta : 'dict'
        All auxilary data to be written to fits header.
    column_names : 'list' of 'string', optional
        names of the fits table columns.
    """
    os.makedirs(path, exist_ok=True)
    mask = spectral_dataset.mask

    hdr = fits.Header()
    for key, value in header_meta.items():
        hdr[key] = value
    primary_hdu = fits.PrimaryHDU(header=hdr)

    columns_list = \
        [fits.Column(name=column_names[0], format='D',
                     unit=spectral_dataset.wavelength_unit.to_string(),
                     array=spectral_dataset.wavelength.data.value[~mask]),
         fits.Column(name=column_names[1], format='D',
                     unit=spectral_dataset.data_unit.to_string(),
                     array=spectral_dataset.data.data.value[~mask]),
         fits.Column(name=column_names[2], format='D',
                     unit=spectral_dataset.data_unit.to_string(),
                     array=spectral_dataset.uncertainty.data.value[~mask]),
        ]
    try:
        columns_list = columns_list + \
            [fits.Column(name="Bin Size", format='D',
                         unit=spectral_dataset.wavelength_binsize_unit.to_string(),
                         array=spectral_dataset.wavelength_binsize.data.value[~mask])]
    except AttributeError:
        pass

    hdu = fits.BinTableHDU.from_columns(columns_list)

    hdul = fits.HDUList([primary_hdu, hdu])
    hdul.writeto(os.path.join(path, filename), overwrite=True)


def write_dataset_to_fits(spectral_dataset, path, filename,
                                      header_meta):
    """
    Write a spectral dataset to a single fits files.

    Parameters
    ----------
    spectral_dataset : 'SpectralDataTimeSeries'
        The data cube which will be save to fits file.
    path : 'str'
        Path to the directory where the fits files will be saved.
    filename: 'str' (optional)
        file name of save fits file.
    header_meta : 'dict'
        All auxilary data to be written to fits header.
    """
    data = spectral_dataset.return_masked_array('data')
    mask = data.mask
    wave = spectral_dataset.return_masked_array('wavelength')
    unc = spectral_dataset.return_masked_array('uncertainty')
    time = spectral_dataset.return_masked_array('time')
    try:
        position = spectral_dataset.return_masked_array('position')
    except:
        position = None
    try:
        scaling = spectral_dataset.return_masked_array('scaling')
    except:
        scaling = None


    hdr = fits.Header()
    for key, value in header_meta.items():
        hdr[key] = value
    primary_hdu = fits.PrimaryHDU(header=hdr)

    hdu_sys = fits.ImageHDU(data.data, name='DATA')
    hdu_sys.header['UNITS'] = spectral_dataset.data_unit.to_string()
    hdu_mask = fits.ImageHDU(np.array(mask, dtype=int), name='MASK')
    hdu_wave = fits.ImageHDU(wave.data, name='WAVELENGTH')
    hdu_wave.header['UNITS'] = spectral_dataset.wavelength_unit.to_string()
    hdu_unc = fits.ImageHDU(unc.data, name='UNCERTAINTY')
    hdu_unc.header['UNITS'] = spectral_dataset.data_unit.to_string()
    hdu_time = fits.ImageHDU(time.data, name='TIME')
    hdu_time.header['UNITS'] = spectral_dataset.time_unit.to_string()

    product_list = [primary_hdu, hdu_sys, hdu_mask, hdu_wave, hdu_unc, hdu_time]

    if position is not None:
        hdu_position = fits.ImageHDU(position.data, name='POSITION')
        hdu_position.header['UNITS'] = spectral_dataset.position_unit.to_string()
        product_list += [hdu_position]
    if scaling is not None:
        hdu_scaling = fits.ImageHDU(scaling.data, name='SCALING')
        hdu_scaling.header['UNITS'] = spectral_dataset.scaling_unit.to_string()
        product_list += [hdu_scaling]

    hdul = fits.HDUList(product_list)

    os.makedirs(path, exist_ok=True)
    hdul.writeto(os.path.join(path, filename), overwrite=True)


def read_dataset_from_fits(path, filename, auxilary_meta):
    """
    Read in a spectral dataset from a single fits files.

    Parameters
    ----------
    path : 'str'
        Path to the directory where the fits file is located.
    filename: 'str' (optional)
        file name of save fits file.
    auxilary_meta : 'list'
        All auxilary data to read from fits header.

    Returns
    -------
    spectral_dataset : 'SpectralDataTimeSeries'
        The data cube read from disk.
    """
    with fits.open(os.path.join(path, filename)) as hdul:
        header_meta = {}
        for key in auxilary_meta:
            value=hdul['PRIMARY'].header[key]
            header_meta[key] = value
        data = np.array(hdul['DATA'].data, dtype=np.float64)
        data_unit = u.Unit(hdul['DATA'].header['UNITS'])
        mask = np.array(hdul['MASK'].data, dtype=bool)
        wavelength =  np.array(hdul['WAVELENGTH'].data, dtype=np.float64)
        wavelength_unit = u.Unit(hdul['WAVELENGTH'].header['UNITS'])
        uncertainty = np.array(hdul['UNCERTAINTY'].data, dtype=np.float64)
        time = np.array(hdul['TIME'].data, dtype=np.float64)
        time_unit = u.Unit(hdul['TIME'].header['UNITS'])
        try:
            position = np.array(hdul['POSITION'].data, dtype=np.float64)
            position_unit = u.Unit(hdul['POSITION'].header['UNITS'])
        except:
            position = None
        try:
            scaling = np.array(hdul['SCALING'].data, dtype=np.float64)
            scaling_unit = u.Unit(hdul['SCALING'].header['UNITS'])
        except:
            scaling = None

    spectral_dataset = SpectralDataTimeSeries(
        wavelength=wavelength,
        wavelength_unit=wavelength_unit,
        data=data,
        data_unit=data_unit,
        uncertainty=uncertainty,
        time=time,
        time_unit=time_unit,
        mask=mask)
    spectral_dataset.add_auxilary(**header_meta)

    if position is not None:
        spectral_dataset.add_measurement(position=position,
                                         position_unit=position_unit)
    if scaling is not None:
        spectral_dataset.add_measurement(scaling=scaling,
                                         scaling_unit=scaling_unit)

    return spectral_dataset


def write_timeseries_to_fits(data, path, additional_file_string=None,
                             delete_old_files=False):
    """
    Write spectral timeseries data object to fits files.

    Parameters
    ----------
    data : 'ndarry' or 'SpectralDataTimeSeries'
        The data cube which will be save to fits file. For each time step
        a fits file will be generated.
    path : 'str'
        Path to the directory where the fits files will be saved.
    additional_file_string : 'str' (optional)
        Additional information to be added to the file name.
    """
    os.makedirs(path, exist_ok=True)
    if delete_old_files:
        try:
            dataProduct = data.dataProduct
        except AttributeError:
            warnings.warn("No dataProduct defined. Removing all fits files "
                          "in target directory before writing spectra.")
            dataProduct = ""
        files = glob.glob(os.path.join(path, '*'+dataProduct+'.fits'))
        for f in files:
            try:
                os.remove(f)
            except OSError as e:
                print("Error: %s : %s" % (f, e.strerror))

    ndim = data.data.ndim
    ntime = data.data.shape[-1]
    try:
        dataFiles = data.dataFiles
        dataFiles = [os.path.split(file)[1] for file in dataFiles]
    except AttributeError:
        if ndim == 2:
            fileBase = "spectrum"
        elif ndim == 3:
            fileBase = "image"
        else:
            fileBase = "image_cube"
        if not isinstance(additional_file_string, type(None)):
            fileBase = fileBase + '_' + str(additional_file_string).strip(' ')
        dataFiles = [fileBase+"_{0:0=4d}.fits".format(it)
                     for it in range(ntime)]

    if ndim == 2:
        for itime, fileName in enumerate(dataFiles):
            hdr = fits.Header()
            # makes sure the data is sorted  according to wavelength
            idx = np.argsort(data.wavelength.data.value[..., itime])
            try:
                hdr['TARGET'] = data.target_name
            except AttributeError:
                pass
            hdr['COMMENT'] = "Created by CASCADe pipeline"
            try:
                hdr['PHASE'] = np.ma.mean(data.time[idx, itime]).value
            except np.ma.MaskError:
                pass
            try:
                hdr['TIME_BJD'] = np.ma.mean(data.time_bjd[idx, itime]).value
                hdr['TBJDUNIT'] = data.time_bjd_unit.to_string()
            except AttributeError:
                pass
            try:
                hdr['SCANDIR'] = data.scan_direction[itime]
            except AttributeError:
                pass
            try:
                hdr['SAMPLENR'] = data.sample_number[itime]
            except AttributeError:
                pass
            try:
                hdr['POSITION'] = np.ma.mean(data.position[idx, itime]).value
                hdr['PUNIT'] = data.position_unit.to_string()
            except AttributeError:
                pass
            try:
                hdr['MEDPOS'] = float(data.median_position)
                hdr['MPUNIT'] = data.position_unit.to_string()
            except AttributeError:
                pass
            try:
                hdr['DISP_POS'] = \
                    np.ma.mean(data.dispersion_position[idx, itime]).value
                hdr['DPUNIT'] = data.dispersion_position_unit.to_string()
            except AttributeError:
                pass
            try:
                hdr['SCALE'] = np.ma.mean(data.scale[idx, itime]).value
                hdr['SUNIT'] = data.scale_unit.to_string()
            except AttributeError:
                pass
            try:
                hdr['ANGLE'] = np.ma.mean(data.angle[idx, itime]).value
                hdr['AUNIT'] = data.angle_unit.to_string()
            except AttributeError:
                pass
            primary_hdu = fits.PrimaryHDU(header=hdr)
            hdu = fits.BinTableHDU.from_columns(
                    [fits.Column(name='LAMBDA', format='D',
                                 unit=data.wavelength_unit.to_string(),
                                 array=data.wavelength.data.value[idx, itime]),
                     fits.Column(name='FLUX', format='D',
                                 unit=data.data_unit.to_string(),
                                 array=data.data.data.value[idx, itime]),
                     fits.Column(name='FERROR', format='D',
                                 unit=data.data_unit.to_string(),
                                 array=data.uncertainty.data.value[idx,
                                                                   itime]),
                     fits.Column(name='MASK', format='L',
                                 array=data.mask[idx, itime])
                     ])

            hdul = fits.HDUList([primary_hdu, hdu])
            hdul.writeto(os.path.join(path, fileName), overwrite=True)
    elif ndim > 2:
        raise ValueError("Saving spectral images to fits file not \
                          implemented yet")


def find(pattern, path):
    """
    Return  a list of all data files.

    Parameters
    ----------
    pattern : 'str'
        Pattern used to search for files.
    path : 'str'
        Path to directory to be searched.

    Returns
    -------
    result : 'list' of 'str'
        Sorted list of filenames matching the 'pattern' search
    """
    result = []
    for root, dirs, files in os.walk(path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(os.path.join(root, name))
    return sorted(result)


def get_data_from_fits(data_files, data_list, auxilary_list):
    """
    Read observations from fits.

    This function reads in a list of fits files containing the
    spectral time series data and auxilary information like position and time.

    Parameters
    ----------
    data_files : 'list' of 'str'
        List containing the names of all fits filaes to be read in.
    data_list : 'list' of 'str'
        List containing all the fits data keywords
    auxilary_list : 'list' of 'str'
        List containing all keywords for the auxilary data

    Returns
    -------
    data_dict : 'dict'
        Dictonary containing all spectral time series data
    auxilary_dict : dict'
        Dictonary containing all auxilary data
    """
    ndata = len(data_files)
    data_struct = {'data': [], 'flag': True}
    data_dict = {k: copy.deepcopy(data_struct) for k in data_list}
    auxilary_data_struct = {"data": np.zeros(ndata),
                            "data_unit": u.dimensionless_unscaled,
                            "flag": False}
    auxilary_dict = {k: copy.deepcopy(auxilary_data_struct)
                     for k in auxilary_list}

    for ifile, fits_file in enumerate(tqdm(data_files,
                                           dynamic_ncols=True)):
        with fits.open(fits_file) as hdu_list:
            fits_header = hdu_list[0].header
            for key, value in auxilary_dict.items():
                try:
                    value['data'][ifile] = fits_header[key]
                    value['flag'] = True
                except KeyError:
                    value['flag'] = False
                except ValueError:
                    if 'UNIT' in key:
                        value['data_unit'] = u.Unit(fits_header[key])
            data_table = QTable.read(hdu_list[1])
            for key, value in data_dict.items():
                try:
                    try:
                        value['data'].append(data_table[key].value *
                                             u.Unit(data_table[key].
                                                    unit.to_string()))
                    except AttributeError:
                        value['data'].append(data_table[key])
                    value['flag'] = True
                except KeyError:
                    value['flag'] = False
    return data_dict, auxilary_dict


@nb.vectorize(["float64(int64,int64,int64,int64)",
               "float32(float32, float32, float32, float32)",
              "float64(float64, float64, float64, float64)"])
def overlap(min1, max1, min2, max2):
    """
    Determine the overlap between two intervals.

    Parameters
    ----------
    min1 : 'float'
        minimum interval 1
    max1 : 'float'
        maximum interval 1
    min2 : 'float'
        minimum interval 2
    max2 : 'float'
        maximum interval 2

    Returns
    -------
    overlap : 'float'
        overlapping range
    """
    return max(0, min(max1, max2) - max(min1, min2))


@nb.jit(nopython=True, cache=True)
def define_limits(wave):
    """
    Define the band for each spectroscopic wavelength.

    Parameters
    ----------
    wave : 'ndarray'
         Wavelengths (1D array)

    Returns
    -------
    lr : 'ndarray'
        lower band limits
    ur : 'ndarray'
        upper band limits
    """
    lr = np.empty(wave.shape, dtype=wave.dtype)
    ur = np.empty(wave.shape, dtype=wave.dtype)
    wd = np.diff(wave)*0.5
    ur[0:-1] = wave[0:-1] + wd
    ur[-1] = wave[-1] + wd[-1]
    lr[1:] = wave[1:] - wd
    lr[0] = wave[0] - wd[0]
    return (lr, ur)


@nb.jit(nopython=True, cache=True)
def define_limits2(wave):
    """
    Define the band for each spectroscopic wavelength for timeseries data.

    Parameters
    ----------
    wave : 'ndarray'
         Wavelengths (2D array)

    Returns
    -------
    lr : 'ndarray'
        lower band limits
    ur : 'ndarray'
        upper band limits
    """
    nwave, ntime = wave.shape
    ur = np.empty((nwave, ntime), dtype=wave.dtype)
    lr = np.empty((nwave, ntime), dtype=wave.dtype)
    w = np.empty((nwave), dtype=wave.dtype)
    for it in nb.prange(ntime):
        w[:] = wave[:, it]
        ls, us = define_limits(w)
        ur[:, it] = us
        lr[:, it] = ls
    return (lr, ur)


def _define_band_limits(wave):
    """
    Define the limits of the rebin intervals.

    Parameters
    ----------
    wave : 'ndarray' of 'float'
        Wavelengths (1D or 2D array)

    Returns
    -------
    limits : 'tuple'
        lower and upper band limits

    Raises
    ------
    TypeError
        When the input wavelength array has more than 2 dimensions
    ValueError
        When the wavelength is not defined for each data point
    """
    if isinstance(wave, np.ma.core.MaskedArray):
        waveUse = wave.data
    else:
        waveUse = wave
    if (not np.all(np.isfinite(waveUse))) | np.any(waveUse <= 0.0):
        raise ValueError("Wavelength not defined for each data point. " +
                         "Stopping rebin")
    ndim = wave.ndim
    if ndim == 1:
        return define_limits(waveUse)
    elif ndim == 2:
        return define_limits2(waveUse)
    else:
        raise TypeError("input wavelength array can have at most 2 dimensions")


@nb.jit(nopython=True, cache=True)
def define_weights(lr0, ur0, lr, ur):
    """
    Weight definition for spectra.

    Parameters
    ----------
    lr0 : 'ndarray'
        lower band limits of reference.
    ur0 : 'ndarray'
        Upper band limits of reference.
    lr : 'ndarray'
        lower band limits.
    ur : 'ndarray'
        upper band limits.

    Returns
    -------
    weights : 'ndarray'
        Integration weights.

    """
    nwave = lr.shape[0]
    nwave0 = lr0.shape[0]
    weights = np.zeros((nwave0, nwave), dtype=lr.dtype)
    for it in nb.prange(nwave0):
        wlr0 = lr0[it]
        wur0 = ur0[it]
        weights[it, :] = overlap(wlr0, wur0, lr, ur)
    return weights


@nb.jit(nopython=True, cache=True)
def define_weights2(lr0, ur0, lr, ur):
    """
    Weight definition for spectral timeseries.

    Parameters
    ----------
    lr0 : 'ndarray'
        lower band limits of reference.
    ur0 : 'ndarray'
        Upper band limits of reference.
    lr : 'ndarray'
        lower band limits.
    ur : 'ndarray'
        upper band limits.

    Returns
    -------
    weights : 'ndarray'
        Integration weights.

    """
    nwave, ntime = lr.shape
    nwave0 = lr0.shape[0]
    weights = np.zeros((nwave0, nwave, ntime), dtype=lr.dtype)
    for it in nb.prange(nwave0):
        wlr0 = lr0[it]
        wur0 = ur0[it]
        weights[it, :, :] = overlap(wlr0, wur0, lr.T, ur.T).T
    return weights


def _define_rebin_weights(lr0, ur0, lr, ur):
    """
    Define the weights used in spectral rebin.

    Define the summation weights (i.e. the fractions of the original intervals
    used in the new interval after rebinning)

    Parameters
    ----------
    lr0 : 'ndarray'
        lower limits wavelength band of new wavelength grid
    ur0 : 'ndarray'
        upper limits wavelength band of new wavelength grid
    lr : 'ndarray'
        lower limits
    ur : 'ndarray'
        upper limits

    Returns
    -------
    weights : 'ndarray' of 'float'
        Summation weights used to rebin to new wavelength grid
    Raises
    ------
    TypeError
       When the input wavelength array has more than 2 dimensions.
    """
    ndim = lr.ndim
    if ndim == 1:
        return define_weights(lr0, ur0, lr, ur)
    elif ndim == 2:
        return define_weights2(lr0, ur0, lr, ur)
    else:
        raise TypeError("input wavelength array can have at most 2 dimensions")


def _rebin_spectra(spectra, errors, weights):
    """
    Rebin spectra.

    Parameters
    ----------
    spectra : 'ndarray'
        Input spectral data values
    errors: 'ndarray'
        Input error on spectral data values
    weights: 'ndarray'
        rebin weights

    Returns
    -------
    newSpectra: 'ndarray'
        Output spectra
    newErrors: 'ndarray'
        Output error
    """
    norm = np.sum(weights, axis=1)
    newSpectra = np.sum(weights*spectra, axis=1)/norm
    newErrors = np.sqrt(np.sum(weights**2*errors**2, axis=1)/norm**2)
    return newSpectra, newErrors


def spectres(new_spec_wavs, old_spec_wavs, spec_fluxes, spec_errs=None):
    """
    SpectRes: A fast spectral resampling function.

    Copyright (C) 2017  A. C. Carnall
    Function for resampling spectra (and optionally associated uncertainties)
    onto a new wavelength basis.

    Parameters
    ----------
    new_spec_wavs : numpy.ndarray
        Array containing the new wavelength sampling desired for the spectrum
        or spectra.

    old_spec_wavs : numpy.ndarray
        1D array containing the current wavelength sampling of the spectrum or
        spectra.

    spec_fluxes : numpy.ndarray
        Array containing spectral fluxes at the wavelengths specified in
        old_spec_wavs, last dimension must correspond to the shape of
        old_spec_wavs.
        Extra dimensions before this may be used to include multiple spectra.

    spec_errs : numpy.ndarray (optional)
        Array of the same shape as spec_fluxes containing uncertainties
        associated with each spectral flux value.

    Returns
    -------
    resampled_fluxes : numpy.ndarray
        Array of resampled flux values, first dimension is the same length
        as new_spec_wavs, other dimensions are the same as spec_fluxes

    resampled_errs : numpy.ndarray
        Array of uncertainties associated with fluxes in resampled_fluxes.
        Only returned if spec_errs was specified.
    """
    # Generate arrays of left hand side positions and widths for the
    # old and new bins
    spec_lhs = np.zeros(old_spec_wavs.shape[0])
    spec_widths = np.zeros(old_spec_wavs.shape[0])
    spec_lhs = np.zeros(old_spec_wavs.shape[0])
    spec_lhs[0] = old_spec_wavs[0] - (old_spec_wavs[1] - old_spec_wavs[0])/2
    spec_widths[-1] = (old_spec_wavs[-1] - old_spec_wavs[-2])
    spec_lhs[1:] = (old_spec_wavs[1:] + old_spec_wavs[:-1])/2
    spec_widths[:-1] = spec_lhs[1:] - spec_lhs[:-1]

    filter_lhs = np.zeros(new_spec_wavs.shape[0]+1)
    filter_widths = np.zeros(new_spec_wavs.shape[0])
    filter_lhs[0] = new_spec_wavs[0] - (new_spec_wavs[1] - new_spec_wavs[0])/2
    filter_widths[-1] = (new_spec_wavs[-1] - new_spec_wavs[-2])
    filter_lhs[-1] = new_spec_wavs[-1]+(new_spec_wavs[-1]-new_spec_wavs[-2])/2
    filter_lhs[1:-1] = (new_spec_wavs[1:] + new_spec_wavs[:-1])/2
    filter_widths[:-1] = filter_lhs[1:-1] - filter_lhs[:-2]

    # Check that the range of wavelengths to be resampled_fluxes onto
    # falls within the initial sampling region
    if filter_lhs[0] < spec_lhs[0] or filter_lhs[-1] > spec_lhs[-1]:
        raise ValueError("spectres: The new wavelengths specified must \
                         fall within the range of the old wavelength values.")

    # Generate output arrays to be populated
    resampled_fluxes = np.zeros(spec_fluxes[..., 0].shape+new_spec_wavs.shape)

    if spec_errs is not None:
        if spec_errs.shape != spec_fluxes.shape:
            raise ValueError("If specified, spec_errs must be the same shape \
                             as spec_fluxes.")
        else:
            resampled_fluxes_errs = np.copy(resampled_fluxes)

    start = 0
    stop = 0

    # Calculate the new spectral flux and uncertainty values,
    # loop over the new bins
    for j in range(new_spec_wavs.shape[0]):

        # Find the first old bin which is partially covered by the new bin
        while spec_lhs[start+1] <= filter_lhs[j]:
            start += 1

        # Find the last old bin which is partially covered by the new bin
        while spec_lhs[stop+1] < filter_lhs[j+1]:
            stop += 1

        # If the new bin falls entirely within one old bin the are the same
        # the new flux and new error are the same as for that bin
        if stop == start:

            resampled_fluxes[..., j] = spec_fluxes[..., start]
            if spec_errs is not None:
                resampled_fluxes_errs[..., j] = spec_errs[..., start]

        # Otherwise multiply the first and last old bin widths by P_ij,
        # all the ones in between have P_ij = 1
        else:

            start_factor = (spec_lhs[start+1] - filter_lhs[j]) / \
                (spec_lhs[start+1] - spec_lhs[start])
            end_factor = (filter_lhs[j+1] - spec_lhs[stop]) / \
                (spec_lhs[stop+1] - spec_lhs[stop])

            spec_widths[start] *= start_factor
            spec_widths[stop] *= end_factor

            # Populate the resampled_fluxes spectrum and uncertainty arrays
            resampled_fluxes[..., j] = np.sum(spec_widths[start:stop+1] *
                                              spec_fluxes[..., start:stop+1],
                                              axis=-1) / \
                np.sum(spec_widths[start:stop+1])

            if spec_errs is not None:
                resampled_fluxes_errs[..., j] = \
                    np.sqrt(np.sum((spec_widths[start:stop+1] *
                                    spec_errs[..., start:stop+1])**2, axis=-1)
                            )/np.sum(spec_widths[start:stop+1])

            # Put back the old bin widths to their initial values for later use
            spec_widths[start] /= start_factor
            spec_widths[stop] /= end_factor

    # If errors were supplied return the resampled_fluxes spectrum and
    # error arrays
    if spec_errs is not None:
        return resampled_fluxes, resampled_fluxes_errs

    # Otherwise just return the resampled_fluxes spectrum array
    else:
        return resampled_fluxes



# -*- coding: utf-8 -*-
import cascade
from cascade.cpm_model import ols
from cascade.cpm_model import return_lambda_grid
import unittest
import numpy as np
from scipy.stats import norm as norm_stats


class TestCpmModel(unittest.TestCase):
    def setUp(self):
        # create linear system with RV
        n = 1000
        x1 = norm_stats.rvs(0, 1, size=n)
        x2 = norm_stats.rvs(0, 1, size=n)
        x3 = norm_stats.rvs(0, 1, size=n)
        self.answer = np.array([10.0, 40.0, 0.1])
        self.A = np.column_stack([x1, x2, x3])
        self.b = self.answer[0] * x1 + self.answer[1] * x2 + \
            self.answer[2] * x3

    def tearDown(self):
        del self.answer
        del self.A
        del self.b

    def test_basic_cpm(self):
        # solve linear Eq.
        (P, Perr, _) = ols(self.A, self.b)
        for i, (result, error) in enumerate(zip(P, Perr)):
            self.assertAlmostEqual(result, self.answer[i], places=None,
                                   msg=None, delta=1.e4*error)


if __name__ == '__main__':
    #  unittest.main()
    suite = unittest.TestLoader().loadTestsFromTestCase(TestCpmModel)
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite)
